/*
 * $Header: /home/cvspublic/jakarta-slide/src/share/org/apache/slide/macro/MacroImpl.java,v 1.16 2001/05/16 12:05:15 juergen Exp $
 * $Revision: 1.16 $
 * $Date: 2001/05/16 12:05:15 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.macro;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;
import java.util.Stack;
import org.apache.slide.common.*;
import org.apache.slide.structure.*;
import org.apache.slide.security.*;
import org.apache.slide.content.*;
import org.apache.slide.lock.*;

/**
 * Macro helper class.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.16 $
 */
public final class MacroImpl implements Macro {
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     *
     * @param namespace Namespace
     * @param namespaceConfig Namespace configuration
     * @param contentHelper Access to content
     * @param structureHelper Access to structure
     */
    public MacroImpl(Namespace namespace, NamespaceConfig namespaceConfig,
                     Security securityHelper, Content contentHelper,
                     Structure structureHelper,
                     Lock lockHelper) {
        this.namespace       = namespace;
        this.namespaceConfig = namespaceConfig;
        this.securityHelper  = securityHelper;
        this.contentHelper   = contentHelper;
        this.structureHelper = structureHelper;
        this.lockHelper      = lockHelper;
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Namespace.
     */
    private Namespace namespace;
    
    
    /**
     * Namespace configuration.
     */
    private NamespaceConfig namespaceConfig;
    
    
    /**
     * Content helper.
     */
    private Content contentHelper;
    
    
    /**
     * Structure helper.
     */
    private Structure structureHelper;
    
    
    /**
     * Security helper.
     */
    private Security securityHelper;
    
    
    /**
     * Lock helper.
     */
    private Lock lockHelper;
    
    
    // ---------------------------------------------------------- Macro Methods
    
    
    /**
     * Recursive copy with overwrite macro.
     *
     * @param token Credentials token
     * @param sourceUri Uri of the source
     * @param destinationUri Uri of the destination
     * @exception CopyMacroException Generic Slide exception
     */
    public void copy(SlideToken token, String sourceUri,
                     String destinationUri)
        throws CopyMacroException, DeleteMacroException {
        this.copy(token, sourceUri, destinationUri,
                  RECURSIVE_OVERWRITE_PARAMETERS);
    }
    
    
    /**
     * Copy macro.
     *
     * @param token Credentials token
     * @param sourceUri Uri of the source
     * @param destinationUri Uri of the destination
     * @param parameters Macro parameters
     * @exception CopyMacroException Generic Slide exception
     */
    public void copy(SlideToken token, String sourceUri,
                     String destinationUri, MacroParameters parameters)
        throws CopyMacroException, DeleteMacroException {
        
        Domain.debug("Copy " + sourceUri + " to " + destinationUri);
        
        // handle the case that source and destination are identical
        if (sourceUri.equals(destinationUri)) {
            CopyMacroException e = new CopyMacroException("Copy failed");
            e.addException(new ForbiddenException(sourceUri));
            throw e;
        }
        
        
        if (parameters.isOverwrite()) {
            try {
                // We make sure the object we want to overwrite exists
                structureHelper.retrieve(token, destinationUri);
                delete(token, destinationUri);
            } catch(ObjectNotFoundException e) {
                // Silent catch, the target doesn't exist
            } catch(SlideException e) {
                // We try to delete the target anyway.
                // It will probably fail again, but we will get a properly
                // packaged exception
                delete(token, destinationUri);
            }
        }
        
        // Now performing the actual copy
        CopyMacroException e = new CopyMacroException("Copy failed");
        
        copyObject(token, sourceUri, destinationUri, parameters, e);
        
        // If there were errors, we throw the nested exception
        if (!e.isEmpty()) {
            throw e;
        }
        
    }
    
    
    /**
     * Recursive move with overwrite macro.
     *
     * @param token Credentials token
     * @param sourceUri Uri of the source
     * @param destinationUri Uri of the destination
     * @exception CopyMacroException Exception occured during copy
     * @exception DeleteMacroException Exception occured during deletion
     */
    public void move(SlideToken token, String sourceUri,
                     String destinationUri)
        throws CopyMacroException, DeleteMacroException {
        move(token, sourceUri, destinationUri,
             RECURSIVE_OVERWRITE_PARAMETERS);
    }
    
    
    /**
     * Move macro.
     *
     * @param token Credentials token
     * @param sourceUri Uri of the source
     * @param destinationUri Uri of the destination
     * @param parameters Macro parameters
     * @exception CopyMacroException Exception occured during copy
     * @exception DeleteMacroException Exception occured during deletion
     */
    public void move(SlideToken token, String sourceUri,
                     String destinationUri, MacroParameters parameters)
        throws CopyMacroException, DeleteMacroException {
        copy(token, sourceUri, destinationUri, parameters);
        delete(token, sourceUri, parameters);
    }
    
    
    /**
     * Recursive delete.
     *
     * @param token Credentials token
     * @param targetUri Uri of the object to delete
     * @exception DeleteMacroException Generic Slide exception
     */
    public void delete(SlideToken token, String targetUri)
        throws DeleteMacroException {
        delete(token, targetUri, RECURSIVE_OVERWRITE_PARAMETERS);
    }
    
    
    /**
     * Delete macro.
     *
     * @param token Credentials token
     * @param targetUri Uri of the source
     * @param parameters Macro parameters, not used right now,
     * so it can be null
     * @exception DeleteMacroException Generic Slide exception
     */
    public void delete(SlideToken token, String targetUri,
                       MacroParameters parameters)
        throws DeleteMacroException {
        Domain.debug("Delete " + targetUri);
        
        DeleteMacroException e = new DeleteMacroException("Delete failed");
        
        deleteObject(token, targetUri, e);
        
        // If there were errors, we throw the nested exception
        if (!e.isEmpty()) {
            throw e;
        }
        
    }
    
    
    // -------------------------------------------------------- Private Methods
    
    
    /**
     * Copy object.
     *
     * @param token Credentials token
     * @param sourceUri Uri of the source
     * @param destinationUri Uri of the destination
     * @param parameters Macro parameters
     * @param CopyMacroException Exception occured during copy
     */
    private void copyObject(SlideToken token, String sourceUri,
                            String destinationUri, MacroParameters parameters,
                            CopyMacroException e) {
        
        Domain.debug("Copy object : from " + sourceUri + " to "
                         + destinationUri);
        
        try {
            
            Vector childrenListVector = new Vector();
            
            // Copying structure
            ObjectNode object = structureHelper.retrieve(token, sourceUri);
            
            // Creating the copy
            try {
                structureHelper.create(token, object.copyObject(), destinationUri);
            }
            catch (ObjectNotFoundException s){
                throw new ConflictException(s.getObjectUri());
            }
            
            // Trying to recreate permissions
            try {
                
                Enumeration permissions = securityHelper
                    .enumeratePermissions(token, object);
                
                while (permissions.hasMoreElements()) {
                    NodePermission permission =
                        (NodePermission) permissions.nextElement();
                    NodePermission newPermission =
                        new NodePermission(destinationUri,
                                           permission.getSubjectUri(),
                                           permission.getActionUri(),
                                           permission.isInheritable());
                    securityHelper.grantPermission(token, newPermission);
                }
                
            } catch (AccessDeniedException ex) {
                // Means that we don't have modifyPermissions rights
                // (root access) on the target.
                // The copy should definitely succeed anyway,
                // so we silently catch the exception.
            }
            
            // Now copying revision descriptors and content
            
            NodeRevisionDescriptors revisionDescriptors =
                contentHelper.retrieve(token, sourceUri);
            
            if (revisionDescriptors.hasRevisions()) {
                
                // Iterators
                NodeRevisionNumber currentRevisionNumber =
                    revisionDescriptors.getInitialRevision();
                NodeRevisionDescriptor currentRevisionDescriptor = null;
                NodeRevisionContent currentRevisionContent = null;
                
                Stack revisionNumbers = new Stack();
                revisionNumbers.push(currentRevisionNumber);
                
                // Creating the initial revision
                currentRevisionDescriptor = contentHelper
                    .retrieve(token, revisionDescriptors,
                              currentRevisionNumber);
                if (currentRevisionDescriptor.getContentLength() > 0) {
                    currentRevisionContent = contentHelper
                        .retrieve(token, revisionDescriptors,
                                  currentRevisionDescriptor);
                } else {
                    currentRevisionContent = null;
                }
                contentHelper.create(token, destinationUri,
                                     currentRevisionDescriptor,
                                     currentRevisionContent);
                
            }
            
            // TODO : Parse stack
            // Algorithm :
            // - While Stack is not empty
            // - Pop a revision number
            // - For each successor of that number, create a new revision
            //   based of the old revision
            // - Add the successor to the stack
            
            // We copy each of this object's children
            
            object = structureHelper.retrieve(token, sourceUri);
            
            Enumeration childrenList = object.enumerateChildren();
            while(childrenList.hasMoreElements()) {
                String childUri = (String) childrenList.nextElement();
                String childDestinationUri = destinationUri + childUri
                    .substring(object.getUri().length());
                copyObject(token, childUri, childDestinationUri,
                           parameters, e);
            }
            
        } catch(SlideException ex) {
            // ex.printStackTrace(); the exception is packed and (hopefully) reported in the response
            e.addException(ex);
        }
        
    }
    
    
    /**
     * Delete object function. Recursive for now.
     *
     * @param token Credentials token
     * @param targetUri Uri of the source
     * @param parameters Macro parameters, not used right now, so it can
     * be null
     * @param e Nested exception
     */
    private void deleteObject(SlideToken token, String targetUri,
                              MacroException e) {
        
        Domain.debug("Delete object : " + targetUri);
        
        try {
            
            ObjectNode currentObject = structureHelper.retrieve(token,
                                                                targetUri);
            
            // Removing children objects
            if (currentObject.hasChildren()) {
                Enumeration children = currentObject.enumerateChildren();
                while (children.hasMoreElements()) {
                    deleteObject(token, (String) children.nextElement(), e);
                }
            }
            
            
            // Removing all revisions
            NodeRevisionDescriptors revisionDescriptors =
                contentHelper.retrieve(token, targetUri);
            
            // if the object is not revisioned remove it completely
            // from it's associated stores
            if (!revisionDescriptors.isVersioned()) {
                // remove the associated locks
                Enumeration locks = lockHelper.enumerateLocks(token, targetUri);
                while (locks.hasMoreElements()) {
                    lockHelper.unlock(token, (NodeLock) locks.nextElement());
                }
                // remove the associated security
                Enumeration permissions = securityHelper.enumeratePermissions(token, targetUri);
                while (permissions.hasMoreElements()) {
                    NodePermission permission = (NodePermission) permissions.nextElement();
                    securityHelper.revokePermission(token, permission);
                }
                // remove the NodeRevisionDescriptor and associated content
                if (revisionDescriptors.hasRevisions()) {
                    contentHelper.remove(token, targetUri,
                                         revisionDescriptors.getInitialRevision());
                }
                
                // remove the NodeRevisionDescriptors object
                contentHelper.remove(token, revisionDescriptors);
                
            }
            
            // Removing object.
            structureHelper.remove(token, currentObject);
            
            
        } catch (SlideException ex) {
            e.addException(ex);
        }
        
    }
    
}



