/*
 * $Header: /home/cvspublic/jakarta-slide/src/share/org/apache/slide/store/AbstractStore.java,v 1.15 2001/05/16 12:06:42 juergen Exp $
 * $Revision: 1.15 $
 * $Date: 2001/05/16 12:06:42 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.store;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.io.FileWriter;
import java.io.IOException;
import javax.transaction.TransactionManager;
import javax.transaction.Transaction;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.XAException;
import javax.transaction.xa.Xid;
import org.apache.slide.common.*;
import org.apache.slide.structure.*;
import org.apache.slide.security.*;
import org.apache.slide.lock.*;
import org.apache.slide.content.*;
import org.apache.slide.util.Messages;
import org.apache.slide.util.ObjectCache;
import org.apache.slide.util.HashMapObjectCache;
import org.apache.slide.transaction.SlideTransactionManager;
import org.apache.slide.transaction.SlideTransaction;

/**
 * Abstract implementation of a store. Handles all caching operations.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.15 $
 */
public abstract class AbstractStore extends AbstractSimpleService
    implements Store {
    
    
    // ----------------------------------------------------------- Constructors
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Default store classname.
     */
    protected String defaultStore =
        "slidestore.reference.MemoryDescriptorsStore";
    
    
    /**
     * Default store instance.
     */
    protected Service defaultStoreInstance;
    
    
    /**
     * Node store.
     */
    protected NodeStore nodeStore;
    
    
    /**
     * Security store.
     */
    protected SecurityStore securityStore;
    
    
    /**
     * Lock store.
     */
    protected LockStore lockStore;
    
    
    /**
     * Revision descriptors store.
     */
    protected RevisionDescriptorsStore revisionDescriptorsStore;
    
    
    /**
     * Revision descriptor store.
     */
    protected RevisionDescriptorStore revisionDescriptorStore;
    
    
    /**
     * Content store.
     */
    protected ContentStore contentStore;
    
    
    /**
     * Active resource manager list.
     */
    protected Service resourceManagers[] = new Service[0];
    

    
    // the name of this store as specified in domain.xml
    private String name;
    
    /**
     * Set the name of the store as specified in domain.xml.
     */
    public void setName(String name) {
        this.name = name;
    }
    
    
    
    /**
     * Return the name of the store as specified in domain.xml.
     */
    public String getName() {
        return this.name;
    }

        
        
    /**
     * Set the scope of the store as specified in domain.xml.
     */
    public void setScope(Scope scope) {
        super.setScope(scope);
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].setScope(scope);
        }
    }
    
    
    
    // ---------------------------------------------------- ServiceImpl Methods
    
    
    /**
     * Namespace setter.
     */
    public void setNamespace(Namespace namespace) {
        
        super.setNamespace(namespace);
        
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].setNamespace(namespace);
        }
        
    }
    
 
        private Hashtable parameters = null;
        
    /**
     * Remeber the store parameters to initialise the default child stores on request
     *
     * @param parameters Hashtable containing the parameters' name
     * and associated value
     * @exception ServiceParameterErrorException Incorrect service parameter
     * @exception ServiceParameterMissingException Service parameter missing
     */
    public void setParameters(Hashtable parameters)
        throws ServiceParameterErrorException,
        ServiceParameterMissingException {
        this.parameters = parameters;
    }
    
    
    /**
     * Connects to descriptors store.
     *
     * @exception DataException
     */
    public void connect()
        throws ServiceConnectionFailedException {
        
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].connect();
        }
        
    }
    
    
    /**
     * Disconnects from descriptors store.
     *
     * @exception ServiceDisconnectionFailedException
     */
    public void disconnect()
        throws ServiceDisconnectionFailedException {
        
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].disconnect();
        }
        
    }
    
    
    /**
     * Initializes descriptors store.
     *
     * @exception ServiceInitializationFailedException Throws an exception
     * if the descriptors store has already been initialized before
     */
    public void initialize(NamespaceAccessToken token)
        throws ServiceInitializationFailedException {
        
        super.initialize(token);
        
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].initialize(token);
        }
        
    }
    
    
    /**
     * Deletes descriptors store. Should remove stored data if possible.
     *
     * @exception ServiceResetFailedException Reset failed
     */
    public void reset()
        throws ServiceResetFailedException {
        
        for (int i = 0; i < resourceManagers.length; i++) {
            resourceManagers[i].reset();
        }
        
    }
    
    
    /**
     * This function tells whether or not the descriptors store is connected.
     *
     * @return boolean true if we are connected
     * @exception ServiceAccessException Service access error
     */
    public boolean isConnected()
        throws ServiceAccessException {
        
        for (int i = 0; i < resourceManagers.length; i++) {
            if (!resourceManagers[i].isConnected())
                return false;
        }
        return true;
        
    }
    
    
    // ----------------------------------------------------- XAResource Methods
    
    
    /**
     * Commit the global transaction specified by xid.
     *
     * @param xid A global transaction identifier
     * @param onePhase If true, the resource manager should use a one-phase
     * commit protocol to commit the work done on behalf of xid.
     * @exception XAException An error has occurred. Possible XAExceptions
     * are XA_HEURHAZ, XA_HEURCOM, XA_HEURRB, XA_HEURMIX, XAER_RMERR,
     * XAER_RMFAIL, XAER_NOTA, XAER_INVAL, or XAER_PROTO. If the resource
     * manager did not commit the transaction and the paramether onePhase is
     * set to true, the resource manager may throw one of the XA_RB*
     * exceptions. Upon return, the resource manager has rolled back the
     * branch's work and has released all held resources.
     */
    public void commit(Xid xid, boolean onePhase)
        throws XAException {
        super.commit(xid, onePhase);
    }

    
    /**
     * Ends the work performed on behalf of a transaction branch.
     *
     * @param xid A global transaction identifier that is the same as what
     * was used previously in the start method.
     * @param flags One of TMSUCCESS, TMFAIL, or TMSUSPEND
     * @exception XAException An error has occurred. Possible XAException
     * values are XAER_RMERR, XAER_RMFAILED, XAER_NOTA, XAER_INVAL,
     * XAER_PROTO, or XA_RB*.
     */
    public void end(Xid xid, int flags)
        throws XAException {
        super.end(xid, flags);
    }
    
    
    /**
     * Tell the resource manager to forget about a heuristically completed
     * transaction branch.
     *
     * @param xid A global transaction identifier
     * @exception XAException An error has occurred. Possible exception values
     * are XAER_RMERR, XAER_RMFAIL, XAER_NOTA, XAER_INVAL, or XAER_PROTO.
     */
    public void forget(Xid xid)
        throws XAException {
        super.forget(xid);
    }
    
    
    /**
     * Ask the resource manager to prepare for a transaction commit of the
     * transaction specified in xid.
     *
     * @param xid A global transaction identifier
     * @return A value indicating the resource manager's vote on the outcome
     * of the transaction. The possible values are: XA_RDONLY or XA_OK. If
     * the resource manager wants to roll back the transaction, it should do
     * so by raising an appropriate XAException in the prepare method.
     * @exception XAException An error has occurred. Possible exception
     * values are: XA_RB*, XAER_RMERR, XAER_RMFAIL, XAER_NOTA, XAER_INVAL,
     * or XAER_PROTO.
     */
    public int prepare(Xid xid)
        throws XAException {
        return super.prepare(xid);
    }
    
    
    /**
     * Inform the resource manager to roll back work done on behalf of a
     * transaction branch.
     *
     * @param xid A global transaction identifier
     * @exception XAException An error has occurred
     */
    public void rollback(Xid xid)
        throws XAException {
        super.rollback(xid);
    }
    
    
    /**
     * Start work on behalf of a transaction branch specified in xid.
     *
     * @param xid A global transaction identifier to be associated with the
     * resource
     * @param flags One of TMNOFLAGS, TMJOIN, or TMRESUME
     * @exception XAException An error has occurred. Possible exceptions are
     * XA_RB*, XAER_RMERR, XAER_RMFAIL, XAER_DUPID, XAER_OUTSIDE, XAER_NOTA,
     * XAER_INVAL, or XAER_PROTO.
     */
    public void start(Xid xid, int flags)
        throws XAException {
        super.start(xid, flags);
    }
    
    
    // ---------------------------------------------------------- Store Methods
    
    
    /**
     * Set the node store associated with this store.
     */
    public void setNodeStore(NodeStore nodeStore) {
        this.nodeStore = (nodeStore != null) ?
            (nodeStore) :
            ((NodeStore) getDefaultStoreInstance());
        addResourceManager(this.nodeStore);
    }
    
    
    /**
     * Set the security store associated with this store.
     */
    public void setSecurityStore(SecurityStore securityStore) {
        this.securityStore = (securityStore != null) ?
            (securityStore) :
            ((SecurityStore) getDefaultStoreInstance());
        addResourceManager(this.securityStore);
    }
    
    
    /**
     * Set the lock store associated with this store.
     */
    public void setLockStore(LockStore lockStore) {
        this.lockStore = (lockStore != null) ?
            (lockStore) :
            ((LockStore) getDefaultStoreInstance());
        addResourceManager(this.lockStore);
    }
    
    
    /**
     * Set the revision descriptors store associated with this store.
     */
    public void setRevisionDescriptorsStore
        (RevisionDescriptorsStore revisionDescriptorsStore) {
        this.revisionDescriptorsStore = (revisionDescriptorsStore != null) ?
            (revisionDescriptorsStore) :
            ((RevisionDescriptorsStore) getDefaultStoreInstance());
        addResourceManager(this.revisionDescriptorsStore);
    }
    
    
    /**
     * Set the revision descriptor store associated with this store.
     */
    public void setRevisionDescriptorStore
        (RevisionDescriptorStore revisionDescriptorStore) {
        this.revisionDescriptorStore = (revisionDescriptorStore != null) ?
            (revisionDescriptorStore) :
            ((RevisionDescriptorStore) getDefaultStoreInstance());
        addResourceManager(this.revisionDescriptorStore);
    }
    
    
    /**
     * Set the content store associated with this store.
     */
    public void setContentStore(ContentStore contentStore) {
        this.contentStore = (contentStore != null) ?
            (contentStore) :
            ((ContentStore) getDefaultStoreInstance());
        addResourceManager(this.contentStore);
    }
    
    
    /**
     * Retrive an object from the Descriptors Store.
     *
     * @param uri Uri of the object we want to retrieve
     * @exception ServiceAccessException Error accessing the Descriptors Store
     * @exception ObjectNotFoundException The object to retrieve was not found
     */
    public ObjectNode retrieveObject(Uri uri)
        throws ServiceAccessException, ObjectNotFoundException {
        ObjectNode objectNode = nodeStore.retrieveObject(uri);
        objectNode.validate(uri.toString());
        return objectNode;
    }
    
    
    /**
     * Store an object in the Descriptors Store.
     *
     * @param object Object to update
     * @exception ServiceAccessException Error accessing the Descriptors Store
     * @exception ObjectNotFoundException The object to update was not found
     */
    public void storeObject(Uri uri, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException {
        ObjectNode tempObject = object.cloneObject();
        tempObject.validate(uri.toString());
        enlist(nodeStore);
        try {
            nodeStore.storeObject(uri, tempObject);
        } catch (ServiceAccessException e) {
            delist(nodeStore, false);
            throw e;
        } catch (ObjectNotFoundException e) {
            delist(nodeStore, false);
            throw e;
        }
        delist(nodeStore);
    }
    
    
    /**
     * Create a new object in the Descriptors Store.
     *
     * @param object SlideObject
     * @param uri Uri of the object we want to create
     * @exception ServiceAccessException Error accessing the Descriptors Store
     * @exception ObjectAlreadyExistsException An object already exists
     * at this Uri
     */
    public void createObject(Uri uri, ObjectNode object)
        throws ServiceAccessException, ObjectAlreadyExistsException {
        ObjectNode tempObject = object.cloneObject();
        tempObject.validate(uri.toString());
        enlist(nodeStore);
        try {
            nodeStore.createObject(uri, tempObject);
        } catch (ServiceAccessException e) {
            delist(nodeStore, false);
            throw e;
        } catch (ObjectAlreadyExistsException e) {
            delist(nodeStore, false);
            throw e;
        }
        delist(nodeStore);
    }
    
    
    /**
     * Remove an object from the Descriptors Store.
     *
     * @param object Object to remove
     * @exception ServiceAccessException Error accessing the Descriptors Store
     * @exception ObjectNotFoundException The object to remove was not found
     */
    public void removeObject(Uri uri, ObjectNode object)
        throws ServiceAccessException, ObjectNotFoundException {
        object.validate(uri.toString());
        enlist(nodeStore);
        try {
            nodeStore.removeObject(uri, object);
        } catch (ServiceAccessException e) {
            delist(nodeStore, false);
            throw e;
        } catch (ObjectNotFoundException e) {
            delist(nodeStore, false);
            throw e;
        }
        delist(nodeStore);
    }
    
    
    /**
     * Store an object permissions in the Descriptors Store.
     *
     * @param permission Permission we want to create
     * @exception ServiceAccessException Error accessing the Descriptors Store
     */
    public void grantPermission(Uri uri, NodePermission permission)
        throws ServiceAccessException {
        NodePermission tempPermission = permission.cloneObject();
        tempPermission.validate(uri.toString());
        enlist(securityStore);
        try {
            securityStore.grantPermission(uri, tempPermission);
        } catch (ServiceAccessException e) {
            delist(securityStore, false);
            throw e;
        }
        delist(securityStore);
    }
    
    
    /**
     * Store an object permissions in the Descriptors Store.
     *
     * @param permission Permission we want to create
     * @exception ServiceAccessException Error accessing the Descriptors Store
     */
    public void revokePermission(Uri uri, NodePermission permission)
        throws ServiceAccessException {
        permission.validate(uri.toString());
        enlist(securityStore);
        try {
            securityStore.revokePermission(uri, permission);
        } catch (ServiceAccessException e) {
            delist(securityStore, false);
            throw e;
        }
        delist(securityStore);
    }
    
    
    /**
     * Revoke all the permissions on the object.
     *
     * @param permission Permission we want to create
     * @exception ServiceAccessException Error accessing the Descriptors Store
     */
    public void revokePermissions(Uri uri)
        throws ServiceAccessException {
        enlist(securityStore);
        try {
            securityStore.revokePermissions(uri);
        } catch (ServiceAccessException e) {
            delist(securityStore, false);
            throw e;
        }
        delist(securityStore);
    }
    
    
    /**
     * Read an object permissions.
     *
     * @param permission Permission we want to create
     * @exception ServiceAccessException Error accessing the Descriptors Store
     */
    public Enumeration enumeratePermissions(Uri uri)
        throws ServiceAccessException {
        // TODO : The vectors elements MUST be cloned
        return securityStore.enumeratePermissions(uri);
    }
    
    
    /**
     * Puts a lock on a subject.
     *
     * @param lock Lock token
     * @exception ServiceAccessException Service access error
     */
    public void putLock(Uri uri, NodeLock lock)
        throws ServiceAccessException {
        lock.validate(uri.toString());
        enlist(lockStore);
        try {
            lockStore.putLock(uri, lock);
        } catch (ServiceAccessException e) {
            delist(lockStore, false);
            throw e;
        }
        delist(lockStore);
    }
    
    
    /**
     * Renews a lock.
     *
     * @param lock Token to renew
     * @exception ServiceAccessException Service access error
     * @exception LockTokenNotFoundException Lock token was not found
     */
    public void renewLock(Uri uri, NodeLock lock)
        throws ServiceAccessException, LockTokenNotFoundException {
        lock.validate(uri.toString());
        enlist(lockStore);
        try {
            lockStore.renewLock(uri, lock);
        } catch (ServiceAccessException e) {
            delist(lockStore, false);
            throw e;
        } catch (LockTokenNotFoundException e) {
            delist(lockStore, false);
            throw e;
        }
        delist(lockStore);
    }
    
    
    /**
     * Removes (cancels) a lock.
     *
     * @param lock Token to remove
     * @exception ServiceAccessException Service access error
     * @exception LockTokenNotFoundException Lock token was not found
     */
    public void removeLock(Uri uri, NodeLock lock)
        throws ServiceAccessException, LockTokenNotFoundException {
        lock.validate(uri.toString());
        enlist(lockStore);
        try {
            lockStore.removeLock(uri, lock);
        } catch (ServiceAccessException e) {
            delist(lockStore, false);
            throw e;
        } catch (LockTokenNotFoundException e) {
            delist(lockStore, false);
            throw e;
        }
        delist(lockStore);
    }
    
    
    /**
     * Kills a lock.
     *
     * @param lock Token to remove
     * @exception ServiceAccessException Service access error
     * @exception LockTokenNotFoundException Lock token was not found
     */
    public void killLock(Uri uri, NodeLock lock)
        throws ServiceAccessException, LockTokenNotFoundException {
        lock.validate(uri.toString());
        enlist(lockStore);
        try {
            lockStore.killLock(uri, lock);
        } catch (ServiceAccessException e) {
            delist(lockStore, false);
            throw e;
        } catch (LockTokenNotFoundException e) {
            delist(lockStore, false);
            throw e;
        }
        delist(lockStore);
    }
    
    
    /**
     * Returns the list of locks put on a subject.
     *
     * @param subject Subject
     * @return Enumeration List of locks which have been put on the subject
     * @exception ServiceAccessException Service access error
     */
    public Enumeration enumerateLocks(Uri uri)
        throws ServiceAccessException {
        return lockStore.enumerateLocks(uri);
    }
    
    
    /**
     * Retrieve a revision descriptors.
     *
     * @param uri Uri
     * @exception ServiceAccessException Service access error
     * @exception RevisionDescriptorNotFoundException Revision descriptor
     * was not found
     */
    public NodeRevisionDescriptors retrieveRevisionDescriptors(Uri uri)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        NodeRevisionDescriptors tempDescriptors =
            revisionDescriptorsStore.retrieveRevisionDescriptors(uri);
        tempDescriptors.validate(uri.toString());
        return tempDescriptors;
    }
    
    
    /**
     * Create new revision descriptors.
     *
     * @param uri Uri
     * @param revisionDescriptors Node revision descriptors
     * @exception ServiceAccessException Service access error
     */
    public void createRevisionDescriptors
        (Uri uri, NodeRevisionDescriptors revisionDescriptors)
        throws ServiceAccessException {
        revisionDescriptors.validate(uri.toString());
        enlist(revisionDescriptorsStore);
        try {
            revisionDescriptorsStore.createRevisionDescriptors
                (uri, revisionDescriptors);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorsStore, false);
            throw e;
        }
        delist(revisionDescriptorsStore);
    }
    
    
    /**
     * Update revision descriptors.
     *
     * @param uri Uri
     * @param revisionDescriptors Node revision descriptors
     * @exception ServiceAccessException Service access error
     * @exception RevisionDescriptorNotFoundException Revision descriptor
     * was not found
     */
    public void storeRevisionDescriptors
        (Uri uri, NodeRevisionDescriptors revisionDescriptors)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        revisionDescriptors.validate(uri.toString());
        enlist(revisionDescriptorsStore);
        try {
            revisionDescriptorsStore.storeRevisionDescriptors
                (uri, revisionDescriptors);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorsStore, false);
            throw e;
        } catch (RevisionDescriptorNotFoundException e) {
            delist(revisionDescriptorsStore, false);
            throw e;
        }
        delist(revisionDescriptorsStore);
    }
    
    
    /**
     * Remove revision descriptors.
     *
     * @param uri Uri
     * @exception ServiceAccessException Service access error
     */
    public void removeRevisionDescriptors(Uri uri)
        throws ServiceAccessException {
        enlist(revisionDescriptorsStore);
        try {
            revisionDescriptorsStore.removeRevisionDescriptors(uri);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorsStore, false);
            throw e;
        }
        delist(revisionDescriptorsStore);
    }
    
    
    /**
     * Retrieve revision descriptor.
     *
     * @param Uri uri
     * @param revisionNumber Node revision number
     */
    public NodeRevisionDescriptor retrieveRevisionDescriptor
        (Uri uri, NodeRevisionNumber revisionNumber)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        NodeRevisionDescriptor revisionDescriptor =
            revisionDescriptorStore.retrieveRevisionDescriptor
            (uri, revisionNumber);
        revisionDescriptor.validate();
        return revisionDescriptor;
    }
    
    
    /**
     * Create new revision descriptor.
     *
     * @param uri Uri
     * @param revisionDescriptor Node revision descriptor
     * @exception ServiceAccessException Service access error
     */
    public void createRevisionDescriptor
        (Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException {
        revisionDescriptor.validate();
        enlist(revisionDescriptorStore);
        try {
            revisionDescriptorStore.createRevisionDescriptor
                (uri, revisionDescriptor);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorStore, false);
            throw e;
        }
        delist(revisionDescriptorStore);
    }
    
    
    /**
     * Update revision descriptor.
     *
     * @param uri Uri
     * @param revisionDescriptors Node revision descriptor
     * @exception ServiceAccessException Service access error
     * @exception RevisionDescriptorNotFoundException Revision descriptor
     * was not found
     */
    public void storeRevisionDescriptor
        (Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        revisionDescriptor.validate();
        enlist(revisionDescriptorStore);
        try {
            revisionDescriptorStore.storeRevisionDescriptor
                (uri, revisionDescriptor);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorStore, false);
            throw e;
        } catch (RevisionDescriptorNotFoundException e) {
            delist(revisionDescriptorStore, false);
            throw e;
        }
        delist(revisionDescriptorStore);
    }
    
    
    /**
     * Remove revision descriptor.
     *
     * @param uri Uri
     * @param revisionNumber Revision number
     * @exception ServiceAccessException Service access error
     */
    public void removeRevisionDescriptor(Uri uri, NodeRevisionNumber number)
        throws ServiceAccessException {
        number.validate();
        enlist(revisionDescriptorStore);
        try {
            revisionDescriptorStore.removeRevisionDescriptor(uri, number);
        } catch (ServiceAccessException e) {
            delist(revisionDescriptorStore, false);
            throw e;
        }
        delist(revisionDescriptorStore);
    }
    
    
    /**
     * Retrive revision content.
     *
     * @param uri Uri
     * @param revisionNumber Node revision number
     */
    public NodeRevisionContent retrieveRevisionContent
        (Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException, RevisionNotFoundException {
        NodeRevisionContent revisionContent =
            contentStore.retrieveRevisionContent(uri, revisionDescriptor);
        revisionContent.validate();
        return revisionContent;
    }
    
    
    /**
     * Create a new revision
     *
     * @param uri Uri
     * @param revisionDescriptor Node revision descriptor
     * @param revisionContent Node revision content
     */
    public void createRevisionContent
        (Uri uri, NodeRevisionDescriptor revisionDescriptor,
         NodeRevisionContent revisionContent)
        throws ServiceAccessException, RevisionAlreadyExistException {
        revisionDescriptor.validate();
        revisionContent.validate();
        enlist(contentStore);
        try {
            contentStore.createRevisionContent(uri, revisionDescriptor,
                                               revisionContent);
        } catch (ServiceAccessException e) {
            delist(contentStore, false);
            throw e;
        } catch (RevisionAlreadyExistException e) {
            delist(contentStore, false);
            throw e;
        }
        delist(contentStore);
    }
    
    
    /**
     * Modify the latest revision of an object.
     *
     * @param uri Uri
     * @param revisionDescriptor Node revision descriptor
     * @param revisionContent Node revision content
     */
    public void storeRevisionContent
        (Uri uri, NodeRevisionDescriptor revisionDescriptor,
         NodeRevisionContent revisionContent)
        throws ServiceAccessException, RevisionNotFoundException {
        revisionDescriptor.validate();
        revisionContent.validate();
        enlist(contentStore);
        try {
            contentStore.storeRevisionContent(uri, revisionDescriptor,
                                              revisionContent);
        } catch (ServiceAccessException e) {
            delist(contentStore, false);
            throw e;
        } catch (RevisionNotFoundException e) {
            delist(contentStore, false);
            throw e;
        }
        delist(contentStore);
    }
    
    
    /**
     * Remove revision.
     *
     * @param uri Uri
     * @param revisionNumber Node revision number
     */
    public void removeRevisionContent
        (Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException {
        revisionDescriptor.validate();
        enlist(contentStore);
        try {
            contentStore.removeRevisionContent(uri, revisionDescriptor);
        } catch (ServiceAccessException e) {
            delist(contentStore, false);
            throw e;
        }
        delist(contentStore);
    }


    // ------------------------------------------------------ Protected Methods
    
    
    /**
     * Return the default store instance.
     */
    protected Service getDefaultStoreInstance() {
        
        if (defaultStoreInstance != null)
            return defaultStoreInstance;
        
        try {
            defaultStoreInstance =
                (Service)Class.forName(defaultStore).newInstance();
            defaultStoreInstance.setParameters(this.parameters);
        } catch (Throwable t) {
            // Should never happen ...
            t.printStackTrace();
        }
        
        return defaultStoreInstance;
        
    }
    
    
    /**
     * Add a new resource manager.
     *
     * @param service New resource manager
     */
    protected void addResourceManager(Service service) {
        if (service == null)
            return;
        // First check if the new store isn't already in the array
        for (int i = 0; i < resourceManagers.length; i++) {
            try {
                if (resourceManagers[i].isSameRM(service))
                    return;
            } catch (XAException e) {
            }
        }
        Service results[] = new Service[resourceManagers.length + 1];
        System.arraycopy(resourceManagers, 0, results, 0,
                         resourceManagers.length);
        results[resourceManagers.length] = service;
        resourceManagers = results;
    }
    
    
    /**
     * Enlist the resource manager in the current transaction.
     */
    protected void enlist()
        throws ServiceAccessException {
        enlist(this);
    }
    
    
    /**
     * Enlist the resource manager in the current transaction.
     */
    protected void enlist(Service service)
        throws ServiceAccessException {
        // Note: No exception is thrown
        boolean enlisted = false;
        // FIXME : Enhance retries.
        int nbTry = 0;
        while ((!enlisted) && (nbTry++ < 20)) {
            Transaction transaction = null;
            try {
                transaction =
                    namespace.getTransactionManager().getTransaction();
            } catch (Exception e) {
            }
            if (transaction == null) {
                System.out.println("WARNING: No active transaction");
                return;
            }
            try {
                enlisted = transaction.enlistResource(service);
            } catch (Exception e) {
                // Something went wrong.
                throw new ServiceAccessException(this, e.getMessage());
            }
            if (!enlisted) {
                try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {
                    // Then go on.
                }
            }
        }
        if (!enlisted) {
            String exMessage = Messages.format
                (AbstractStore.class.getName() + ".enlistFail", service);
            setRollbackOnly();
            throw new ServiceAccessException(this, exMessage);
        }
    }
    
    
    /**
     * Delist (suspend) the resource manager in the current transaction.
     */
    protected void delist(boolean success)
        throws ServiceAccessException {
        delist(this, success);
    }
    
    
    /**
     * Delist (suspend) the resource manager in the current transaction.
     */
    protected void delist(Service service)
        throws ServiceAccessException {
        delist(service, true);
    }
    
    
    /**
     * Delist (suspend) the resource manager in the current transaction.
     */
    protected void delist(Service service, boolean success)
        throws ServiceAccessException {
        try {
            Transaction transaction =
                namespace.getTransactionManager().getTransaction();
            if (transaction == null)
                return;
            if (success) {
                transaction.delistResource(service, TMSUSPEND);
            } else {
                transaction.delistResource(service, TMFAIL);
            }
        } catch (Exception e) {
            // Something went wrong.
            throw new ServiceAccessException(this, e.getMessage());
        }
    }
    
    
    /**
     * Mark transaction as rollback in case of enlistment failure.
     */
    protected void setRollbackOnly() {
        try {
            Transaction transaction =
                namespace.getTransactionManager().getTransaction();
            if (transaction == null)
                return;
            transaction.setRollbackOnly();
        } catch (Exception e) {
        }
    }
    
    
}

