/*****************************************************************************
 * Copyright (C) The Apache Software Foundation. All rights reserved.        *
 * ------------------------------------------------------------------------- *
 * This software is published under the terms of the Apache Software License *
 * version 1.1, a copy of which has been included  with this distribution in *
 * the LICENSE file.                                                         *
 *****************************************************************************/

package org.apache.slide.util.conf;

import java.util.*;

/**
 * <code>Configuration</code> is a interface encapsulating a configuration node
 * used to retrieve configuration values. This is a "read only" interface
 * preventing applications from modifying their own configurations.
 * <br />
 *
 * The contract surrounding the <code>Configuration</code> is that once
 * it is created, information never changes.  The <code>Configuration</code>
 * is built by the <code>SAXConfigurationBuilder</code> and the
 * <code>ConfigurationImpl</code> helper classes.
 *
 * @version 1.1.0, 22/05/1999.
 * @author <a href="mailto:scoobie@pop.systemy.it">Federico Barbieri</a>
 * @author <a href="mailto:pier@apache.org">Pierpaolo Fumagalli</a>
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @author <a href="mailto:bloritsch@apache.org">Berin Loritsch</a>
 * @author <a href="http://java.apache.org/">Java Apache Project</a>
 */
public interface Configuration {

    /**
     * Return the name of the node.
     *
     * @post getName() != null
     *
     * @returns name of the <code>Configuration</code> node.
     */
    public String getName();

    /**
     * Return a new <code>Configuration</code> instance encapsulating the
     * specified child node.
     *
     * @pre child != null
     * @post getConfiguration() != null
     * @param child The name of the child node.
     *
     * @returns Configuration
     *
     * @exception ConfigurationException If no child with that name exists.
     */
    public Configuration getConfiguration(String child)
    throws ConfigurationException;

    /**
     * Return an <code>Enumeration</code> of <code>Configuration<code>
     * elements containing all node children with the specified name.
     *
     * @pre name != null
     * @post getConfigurations() != null
     * @param name The name of the children to get.
     *
     * @returns Enumeration.  The <code>Enumeration</code> will be
     *          empty if there are no nodes by the specified name.
     */
    public Enumeration getConfigurations(String name);

    /**
     * Return the value of specified attribute.
     *
     * @pre paramName != null
     * @post getAttribute != null
     * @param paramName The name of the parameter you ask the value of.
     *
     * @returns String value of attribute.
     *
     * @exception ConfigurationException If no attribute with that name exists.
     */
    public String getAttribute(String paramName)
    throws ConfigurationException;

    /**
     * Return the <code>int</code> value of the specified attribute contained
     * in this node.
     *
     * @pre paramName != null
     * @post getAttributeAsInt() != null
     * @param paramName The name of the parameter you ask the value of.
     *
     * @returns int value of attribute
     *
     * @exception ConfigurationException If no parameter with that name exists.
     *                                   or if conversion to <code>int</code> fails.
     */
    public int getAttributeAsInt(String paramName)
    throws ConfigurationException;

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>long</code>.
     *
     * @pre paramName != null
     * @post getAttributeAsLong() != null
     * @param paramName The name of the parameter you ask the value of.
     *
     * @returns long value of attribute
     *
     * @exception ConfigurationException If no parameter with that name exists.
     *                                   or if conversion to <code>long</code> fails.
     */
    public long getAttributeAsLong(String name)
	throws ConfigurationException;

    /**
     * Return the <code>float</code> value of the specified parameter contained
     * in this node.
     *
     * @pre paramName != null
     * @post getAttributeAsFloat() != null
     * @param paramName The name of the parameter you ask the value of.
     *
     * @returns float value of attribute
     *
     * @exception ConfigurationException If no parameter with that name exists.
     *                                   or if conversion to <code>float</code> fails.
     */
    public float getAttributeAsFloat(String paramName)
    throws ConfigurationException;

    /**
     * Return the <code>boolean</code> value of the specified parameter contained
     * in this node.<br>
     *
     * @pre paramName != null
     * @post getAttributeAsBoolean() != null
     * @param paramName The name of the parameter you ask the value of.
     *
     * @returns boolean value of attribute
     *
     * @exception ConfigurationException If no parameter with that name exists.
     *                                   or if conversion to <code>boolean</code> fails.
     */
    public boolean getAttributeAsBoolean(String paramName)
    throws ConfigurationException;

    /**
     * Return the <code>String</code> value of the node.
     *
     * @post getValue() != null
     *
     * @returns the value of the node.
     */
    public String getValue();

    /**
     * Return the <code>int</code> value of the node.
     *
     * @post getValueAsInt() != null
     *
     * @returns the value of the node.
     *
     * @exception ConfigurationException If conversion to <code>int</code> fails.
     */
    public int getValueAsInt()
    throws ConfigurationException;

    /**
     * Return the <code>float</code> value of the node.
     *
     * @post getValueAsFloat() != null
     *
     * @returns the value of the node.
     *
     * @exception ConfigurationException If conversion to <code>float</code> fails.
     */
    public float getValueAsFloat()
    throws ConfigurationException;

    /**
     * Return the <code>boolean</code> value of the node.
     *
     * @post getValueAsBoolean() != null
     *
     * @returns the value of the node.
     *
     * @exception ConfigurationException If conversion to <code>boolean</code> fails.
     */
    public boolean getValueAsBoolean()
    throws ConfigurationException;

    /**
     * Return the <code>long</code> value of the node.<br>
     *
     * @post getValueAsLong() != null
     *
     * @returns the value of the node.
     *
     * @exception ConfigurationException If conversion to <code>long</code> fails.
     */
    public long getValueAsLong()
    throws ConfigurationException;

    /**
     * Returns the value of the configuration element as a <code>String</code>.
     * If the configuration value is not set, the default value will be
     * used.
     *
     * @pre defaultValue != null
     * @post getValue(defaultValue) != null
     * @param defaultValue The default value desired.
     *
     * @returns String value of the <code>Configuration</code>, or default
     *          if none specified.
     */
	public String getValue(String defaultValue);

    /**
     * Returns the value of the configuration element as an <code>int</code>.
     * If the configuration value is not set, the default value will be
     * used.
     *
     * @pre defaultValue != null
     * @post getValueAsInt(defaultValue) != null
     * @param defaultValue The default value desired.
     *
     * @returns int value of the <code>Configuration</code>, or default
     *          if none specified.
     */
    public int getValueAsInt(int defaultValue);

    /**
     * Returns the value of the configuration element as a <code>long</code>.
     * If the configuration value is not set, the default value will be
     * used.
     *
     * @pre defaultValue != null
     * @post getValueAsLong(defaultValue) != null
     * @param defaultValue The default value desired.
     *
     * @returns long value of the <code>Configuration</code>, or default
     *          if none specified.
     */
    public long getValueAsLong(long defaultValue);

    /**
     * Returns the value of the configuration element as a <code>float</code>.
     * If the configuration value is not set, the default value will be
     * used.
     *
     * @pre defaultValue != null
     * @post getValueAsFloat(defaultValue) != null
     * @param defaultValue The default value desired.
     *
     * @returns float value of the <code>Configuration</code>, or default
     *          if none specified.
     */
    public float getValueAsFloat(float defaultValue);

    /**
     * Returns the value of the configuration element as a <code>boolean</code>.
     * If the configuration value is not set, the default value will be
     * used.
     *
     * @pre defaultValue != null
     * @post getValueAsBoolean(defaultValue) != null
     * @param defaultValue The default value desired.
     *
     * @returns boolean value of the <code>Configuration</code>, or default
     *          if none specified.
     */
    public boolean getValueAsBoolean(boolean defaultValue);

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>String</code>, or the default value if no attribute by
     * that name exists or is empty.
     *
     * @pre name != null
     * @pre defaultValue != null
     * @post getAttribute(name, defaultValue) != null
     * @param name The name of the attribute you ask the value of.
     * @param defaultValue The default value desired.
     *
     * @returns String value of attribute. It will return the default
     *          value if the named attribute does not exist, or if
     *          the value is not set.
     */
	public String getAttribute(String name, String defaultValue);

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>int</code>, or the default value if no attribute by
     * that name exists or is empty.
     *
     * @pre name != null
     * @pre defaultValue != null
     * @post getAttributeAsInt(name, defaultValue) != null
     * @param name The name of the attribute you ask the value of.
     * @param defaultValue The default value desired.
     *
     * @returns int value of attribute. It will return the default
     *          value if the named attribute does not exist, or if
     *          the value is not set.
     */
    public int getAttributeAsInt(String name, int defaultValue);

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>long</code>, or the default value if no attribute by
     * that name exists or is empty.
     *
     * @pre name != null
     * @pre defaultValue != null
     * @post getAttributeAsLong(name, defaultValue) != null
     * @param name The name of the attribute you ask the value of.
     * @param defaultValue The default value desired.
     *
     * @returns long value of attribute. It will return the default
     *          value if the named attribute does not exist, or if
     *          the value is not set.
     */
    public long getAttributeAsLong(String name, long defaultValue);

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>float</code>, or the default value if no attribute by
     * that name exists or is empty.
     *
     * @pre name != null
     * @pre defaultValue != null
     * @post getAttributeAsFloat(name, defaultValue) != null
     * @param name The name of the attribute you ask the value of.
     * @param defaultValue The default value desired.
     *
     * @returns float value of attribute. It will return the default
     *          value if the named attribute does not exist, or if
     *          the value is not set.
     */
    public float getAttributeAsFloat(String name, float defaultValue);

    /**
     * Returns the value of the attribute specified by its name as a
     * <code>boolean</code>, or the default value if no attribute by
     * that name exists or is empty.
     *
     * @pre name != null
     * @pre defaultValue != null
     * @post getAttributeAsBoolean(name, defaultValue) != null
     * @param name The name of the attribute you ask the value of.
     * @param defaultValue The default value desired.
     *
     * @returns boolean value of attribute. It will return the default
     *          value if the named attribute does not exist, or if
     *          the value is not set.
     */
    public boolean getAttributeAsBoolean(String name, boolean defaultValue);

	/**
	 * Return a <code>String</code> indicating the position of this
	 * configuration element in a source file or URI.
     *
     * @returns String if a source file or URI is specified.  Otherwise
     *          it returns <code>null</code>
	 */
	public String getLocation();
}
