/*****************************************************************************
 * Copyright (C) The Apache Software Foundation. All rights reserved.        *
 * ------------------------------------------------------------------------- *
 * This software is published under the terms of the Apache Software License *
 * version 1.1, a copy of which has been included  with this distribution in *
 * the LICENSE file.                                                         *
 *****************************************************************************/
 
package org.apache.slide.util.conf;

import java.io.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * @author <a href="mailto:scoobie@betaverion.org">Federico Barbieri</a>
 * @author <a href="mailto:arkin@exoffice.com>">Assaf Arkin</a>
 */
public class Populate implements DocumentHandler, ErrorHandler {
    
    private Element    _root;
    private Element    _current;
    private Locator     _locator;

    public Element load(InputSource is, Parser parser)
    throws SAXException, IOException, ConfigurationException {
        parser.setDocumentHandler(this);
        parser.parse(is);
        return _root;
    }
    
    
    /**
     * Called to reset this object so it can be used a second time.
     */
    public void reset() {
        _root = null;
    }
    
    
    public void startDocument()
    throws SAXParseException {
        // Starting a new document with the same populate object. This object is not
        // reusable with a reset.
        if  ( _root != null ) 
            throw new SAXParseException( "Cannot start processing a new document without a reset", _locator );
    }
    
    
    public void endDocument()
        throws SAXParseException
    {
        // Not all elements have been closed, but end of document has been reached.
        // This should never happen.
        if ( _current != null )
            throw new SAXParseException( "Not all elements have been closed at end of document.", _locator );
    }
    
    
    public void startElement( String name, AttributeList attr ) {
        int     i;
        Element parent;
        
        // If this is the root element, create the first element, else
        // create a child for the current element and process it.
        if ( _current == null ) {
            _current = new Element( name, null );
            _root = _current;
        } else {
            parent = _current;
            _current = new Element( name, parent );
            parent.addChild( _current );
        }
        // Set the element's name and the attributes one by one, so the
        // object tree will be a reflection of the XML document.
        _current.setName( name );
        for ( i = 0 ; i < attr.getLength() ; ++i )
            _current.setAttribute( attr.getName( i ), attr.getValue( i ) );
    }
    
    
    /**
     * Closing the element. This is the place to check special conditions about the
     * element content. If not, just make the parent element the current element.
     */
    public void endElement( String name )
    throws SAXParseException {
        // Attempt to close an element when the root element has already been
        // closed. Should never happen.
        if ( _current == null )
            throw new SAXParseException( "Attempt to close the element " + name +
                " when root element is already closed.", _locator );
        // Attempt to close one element when a different element is open and
        // waiting to be closed. Should never happen.
        if ( ! _current.getName().equals( name ) )
            throw new SAXParseException( "Attempt to close the element " + name +
                " when the element " + _current.getName() + " should be closed.", _locator );
        // All we have to do is go back to the parent.
        _current = _current.getParent();
    }
    
    
    /**
     * Called when there is character data (text) inside an element. Will accumulate
     * all the character data and store it inside the object data.
     */
    public void characters( char[] ch, int start, int length )
        throws SAXParseException {
        Object          data;
        StringBuffer    buf;
        
        if ( ch == null || length == 0 ) return;
        // Attempt to place character before or after the root element.
        // Should never happen.
        if ( _current == null )
            throw new SAXParseException( "Attempt to place character before or after the root element.", _locator );

        // I am assuming that initially data may be just an empty string, or a null.
        // I am assuming that multiple calls to character can occur inside the
        // element and that all the character data should be collected as one.
        data = _current.getData();
        if ( data == null || ! ( data instanceof String ) || ( (String) data ).length() == 0 )
            _current.setData( new String( ch, start, length ) );
        else {
            buf = new StringBuffer( (String) data );
            buf.append( ch, start, length );
            _current.setData( buf.toString() );
        }
    }
     
    
    /**
     * Ignoreable whitespace is just space inside an element that only contains
     * other elements. We can safely ignore it.
     */
    public void ignorableWhitespace( char[] ch, int start, int length ) {
    }
    
    
    public void processingInstruction( String target, String pi ) {
    }
    
    
    /**
     * We can store the locate just to know where an error occurs.
     */
    public void setDocumentLocator( Locator locator ) {
        _locator = locator;
    }
    
    
    public void error( SAXParseException except ) {
        System.out.println( except.getMessage() );
    }

    public void fatalError( SAXParseException except ) {
        System.out.println( except.getMessage() );
    }

    public void warning( SAXParseException except ) {
        System.out.println( except.getMessage() );
    }
}
