/*
 * $Header: /home/cvspublic/jakarta-slide/src/webdav/server/org/apache/slide/webdav/method/WebdavMethod.java,v 1.22 2001/05/29 02:38:25 remm Exp $
 * $Revision: 1.22 $
 * $Date: 2001/05/29 02:38:25 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.webdav.method;

import java.security.Principal;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.*;
import org.apache.util.URLUtil;
import org.apache.util.MD5Encoder;
import org.apache.util.WebdavStatus;
import org.apache.slide.common.*;
import org.apache.slide.authenticate.*;
import org.apache.slide.structure.*;
import org.apache.slide.content.*;
import org.apache.slide.lock.*;
import org.apache.slide.macro.*;
import org.apache.slide.security.*;
import org.apache.slide.webdav.*;

/**
 * WebDAV method.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 */
public abstract class WebdavMethod {
    
    
    // -------------------------------------------------------------- Constants
    
    
    public static final String LOCK_TOKEN = "opaquelocktoken:";
    public static final String SLIDE_NAMESPACE =
        "http://jakarta.apache.org/slide/";
    public static final String SLIDE_NAMESPACE_ABBREV = "S";
    
    public static final String PRINCIPAL = 
        "org.apache.slide.webdav.method.principal";
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Requested Uri.
     */
    protected String requestUri;
    
    
    /**
     * Principal.
     */
    protected Principal principal;
    
    
    /**
     * Servlet request.
     */
    protected HttpServletRequest req;
    
    
    /**
     * Servlet response.
     */
    protected HttpServletResponse resp;
    
    
    /**
     * HTTP servlet.
     */
    protected GenericServlet servlet;
    
    
    /**
     * Servlet context.
     */
    protected ServletContext context;
    
    /**
     * Request body.
     */
    protected String requestBody;
    
    
    /**
     * Namespace access token.
     */
    protected NamespaceAccessToken token;
    
    
    /**
     * Files path.
     */
    protected String filesPath;
    
    
    /**
     * Structure helper.
     */
    protected Structure structure;
    
    
    /**
     * Content helper.
     */
    protected Content content;
    
    
    /**
     * Security helper.
     */
    protected Security security;
    
    
    /**
     * Lock helper.
     */
    protected Lock lock;
    
    
    /**
     * Macro helper.
     */
    protected Macro macro;
    
    
    /**
     * Credentials token associated with the principal.
     */
    protected CredentialsToken credToken;
    
    
    /**
     * Slide token.
     */
    protected SlideToken slideToken;
    
    
    /**
     * DOM XML parser.
     */
    protected DocumentBuilder documentBuilder;
    
    
    /**
     * Namespaces list. Keyed by abbreviation.
     */
    protected Hashtable namespaces;
    
    
    /**
     * Default namespace.
     */
    protected String defaultNamespace = NodeProperty.DEFAULT_NAMESPACE;
    
    
    /**
     * MD5 message digest provider.
     */
    protected static MessageDigest md5Helper;


    /**
     * The MD5 helper object for this class.
     */
    protected static final MD5Encoder md5Encoder = new MD5Encoder();
    
    
    /**
     * Manager servlet path.
     */
    protected static String managerServletPath = "/manager/";
    
    
    /**
     * Depth limit. To limit tree browsing when using depth = infinity
     */
    protected static int depthLimit = 3;


    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Method constructor.
     */
    public WebdavMethod(GenericServlet servlet, NamespaceAccessToken token,
                        HttpServletRequest req, HttpServletResponse resp) {
        
        if (documentBuilder == null) {
            try {
                documentBuilder =
                    DocumentBuilderFactory.newInstance().newDocumentBuilder();
            } catch(ParserConfigurationException e) {
                e.printStackTrace();
            }
        }
        
        this.principal = req.getUserPrincipal();
        this.req = req;
        this.resp = resp;
        this.servlet = servlet;
        this.context = servlet.getServletContext();
        
        this.token = token;
        this.structure = token.getStructureHelper();
        this.content = token.getContentHelper();
        this.security = token.getSecurityHelper();
        this.lock = token.getLockHelper();
        this.macro = token.getMacroHelper();

        if (principal == null) {
            this.principal = 
                (Principal) req.getSession(true).getAttribute(PRINCIPAL);
        } else {
            req.getSession(true).setAttribute(PRINCIPAL, principal);
        }
        
        if (principal == null) {
            this.credToken = new CredentialsToken("");
        } else {
            this.credToken = new CredentialsToken(principal);
        }
        
        this.slideToken = new SlideToken(this.credToken);
        this.slideToken.setEnforceLockTokens(true);
        
        this.requestUri = getRelativePath(req);
        
        parseHeaders();
        
    }
    
    
    // --------------------------------------------------------- Public Methods
    
    
    /**
     * Initializes static fields.
     *
     * @param servletPath Path of the manager servlet
     */
    public static void init(String servletPath, int limit) {
        
        // Load the MD5 helper used to calculate signatures.
        try {
            md5Helper = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
            throw new IllegalStateException();
        }
        
        if (servletPath != null) {
            managerServletPath = servletPath;
        }
        
        if (limit>0) {
            depthLimit = limit;
        }

    }
        
    
    /**
     * Exceute method.
     *
     * @exception WebdavException
     */
    public void run()
        throws WebdavException {
        
        boolean transactionIsStarted = false;
        
        try {
            parseRequest();
            if (methodNeedsTransactionSupport()) {
                token.begin();
                transactionIsStarted = true;
            }
            executeRequest();
            if (methodNeedsTransactionSupport()) {
                token.commit();
                transactionIsStarted = false;
            }
        } catch (WebdavException ex) {
            // ignore the WebDav Exception and assume that the response code is already set.
        } catch (Exception ex) {
            ex.printStackTrace();
            resp.setStatus(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
            throw new WebdavException(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
        } finally {
            if (transactionIsStarted && methodNeedsTransactionSupport()) {
                // something went wrong, we are here and the TA is still open
                try {
                    token.rollback();
                } catch (Exception e) {
                }
            }
        }
        
    }
    
    
    // ------------------------------------------------------ Protected Methods
    
    
    /**
     * Return true if the method needs transaction support.
     * Return true if the method needs transaction support, e.g.
     * the methods put, proppatch would need TA support. The
     * methods get, propfind would not need transaction support.
     */
    protected boolean methodNeedsTransactionSupport() {
        return true;
        // this method should return false and e.g. PutMethod should
        // overwrite it to true, but PropFind creates a RevisionDescriptors
        // on the fly
    }
    
    
    /**
     * Return the relative path associated with this servlet.
     *
     * @param request The servlet request we are processing
     */
    public static String getRelativePath(HttpServletRequest request) {
        
        String result = request.getPathInfo();
        if (result == null) {
            result = request.getServletPath();
        }
        if ((result == null) || (result.equals(""))) {
            result = "/";
        }
        return URLDecode(result);
        
    }
    
    
    /**
     * Return an absolute URL (absolute in the HTTP sense) based on a Slide
     * path.
     */
    public String getFullPath(String path) {
        if (path.startsWith("/"))
            return URLEncode(req.getContextPath() + path);
        else
            return URLEncode(req.getContextPath() + "/" + path);
    }
    
    
    /**
     * Parse a node for namespace declaration attributes.
     *
     * @param node DOM node
     */
    protected void parseNodeNamespaceDeclarations(Node node) {
        
        NamedNodeMap nodeMap = node.getAttributes();
        if (nodeMap != null) {
            for (int i = 0; i < nodeMap.getLength(); i++) {
                Node currentNode = nodeMap.item(i);
                if (currentNode.getNodeType() != Node.ATTRIBUTE_NODE)
                    continue;
                String attributeName = currentNode.getNodeName();
                if (attributeName.startsWith("xmlns")) {
                    // We found a namespace declaration
                    if (attributeName.equals("xmlns")) {
                        defaultNamespace = currentNode.getNodeValue();
                    } else {
                        // Stripping out the prefix
                        if (attributeName.startsWith("xmlns:")) {
                            String namespaceAbbreviation =
                                attributeName.substring(6);
                            namespaces.put(namespaceAbbreviation,
                                           currentNode.getNodeValue());
                        }
                    }
                }
            }
        }
        
    }
    
    
    /**
     * Read request contents.
     *
     * @param req Request object handed out by the servlet container
     * @return char[] Array of char which contains the body of the request
     */
    protected void readRequestContent() {
        
        namespaces = new Hashtable();
        namespaces.put(SLIDE_NAMESPACE_ABBREV, SLIDE_NAMESPACE);
        
        if (req.getContentLength() == 0)
            return;
        
        // TODO : Modify this and make it chunking aware
        
        try {
            requestBody = new String(NodeRevisionContent.readFromStream(req.getInputStream()),
                                     getEncodingString(req.getCharacterEncoding()));
        }
        catch (Exception e) { e.printStackTrace();}

        
    }

    
    
    /**
     * Translate the encoding string into a proper Java encoding String.
     */
    public static String getEncodingString(String httpEncoding) {
        String result = httpEncoding;
        if (result == null) result = System.getProperty("file.encoding");
        if (result.startsWith("\"")) result = result.substring(1, result.length());
        if (result.endsWith("\""))   result = result.substring(0, result.length()-1);
        return result;
    }
    
    
    /**
     * Parse lock token headers.
     */
    protected void parseHeaders() {
        
        // Retrieve if header
        String ifHeader = req.getHeader("If");
        //System.out.println("If header : " + ifHeader);
        
        if (ifHeader == null)
            return;
        
        // Parse if header and extract the lock tokens
        int pos = ifHeader.indexOf(LOCK_TOKEN);
        int endPos = -1;
        int offset = LOCK_TOKEN.length();
        String lockToken = null;
        
        while (pos != -1) {
            
            endPos = ifHeader.indexOf('>', pos + offset);
            if (endPos == -1) {
                lockToken = ifHeader;
            } else {
                lockToken = ifHeader.substring(pos + offset, endPos);
            }
            
            //System.out.println("Lock Token found :-" + lockToken + "-");
            slideToken.addLockToken(lockToken);
            
            pos = ifHeader.indexOf(LOCK_TOKEN, endPos);
            
        }
        
    }
    
    
    /**
     * Returns the lock token value from the lock token header.
     *
     * @return String Opaque lock token value
     */
    protected String parseLockToken(String lockTokenValue) {
        int semi = lockTokenValue.indexOf(":");
        String result = null;
        try {
            result = lockTokenValue.substring(semi + 1,
                                              lockTokenValue.length() - 1);
        } catch (Exception e) {
            e.printStackTrace();
        }
        return result;
    }
    
    
    /**
     * Tests if a resource is a collection.
     */
    protected boolean isCollection(NodeRevisionDescriptor revisionDescriptor) {
        
        boolean result = false;
        
        if (revisionDescriptor == null)
            return true;
        
        NodeProperty property = revisionDescriptor.getProperty("resourcetype");
        
        if ((property != null)
            && (property.getValue().equals("<collection/>"))) {
            result = true;
        }
        
        return result;
        
    }
    /**
     * Parse WebDAV XML query.
     *
     * @exception WebdavException
     */
    protected abstract void parseRequest()
        throws WebdavException, IOException;
    
    
    /**
     * Generate XML response.
     *
     * @exception WebdavException
     */
    protected abstract void executeRequest()
        throws WebdavException, IOException;
    
    
    /**
     * Simulate MS IIS5 ?
     *
     * @return boolean
     */
    protected boolean isMsProprietarySupport() {
        return (token.getNamespaceConfig().getParameter("ms") != null);
    }
    
    
    /**
     * Return a context-relative path, beginning with a "/", that represents
     * the canonical version of the specified path after ".." and "." elements
     * are resolved out.  If the specified path attempts to go outside the
     * boundaries of the current context (i.e. too many ".." path elements
     * are present), return <code>null</code> instead.
     *
     * @param path Path to be normalized
     */
    protected static String URLDecode(String path) {

        if (path == null)
            return null;

        // Resolve encoded characters in the normalized path,
        // which also handles encoded spaces so we can skip that later.
        // Placed at the beginning of the chain so that encoded
        // bad stuff(tm) can be caught by the later checks
        String normalized = path;
        if (normalized.indexOf('%') >= 0)
            // FIXME: Use configurable encoding here.
            normalized = URLUtil.URLDecode(normalized, "UTF8");
        if (normalized == null)
            return (null);
        
        // Normalize the slashes and add leading slash if necessary
        if (normalized.indexOf('\\') >= 0)
            normalized = normalized.replace('\\', '/');
        if (!normalized.startsWith("/"))
            normalized = "/" + normalized;
    
        // Resolve occurrences of "//" in the normalized path
        while (true) {
            int index = normalized.indexOf("//");
            if (index < 0)
            break;
            normalized = normalized.substring(0, index) +
            normalized.substring(index + 1);
        }
    
        // Resolve occurrences of "/./" in the normalized path
        while (true) {
            int index = normalized.indexOf("/./");
            if (index < 0)
            break;
            normalized = normalized.substring(0, index) +
            normalized.substring(index + 2);
        }
    
        // Resolve occurrences of "/../" in the normalized path
        while (true) {
            int index = normalized.indexOf("/../");
            if (index < 0)
            break;
            if (index == 0)
            return (null);  // Trying to go outside our context
            int index2 = normalized.lastIndexOf('/', index - 1);
            normalized = normalized.substring(0, index2) +
            normalized.substring(index + 3);
        }
    
        // Return the normalized path that we have completed
        return (normalized);

    }


    /**
     * URL rewriter.
     *
     * @param path Path which has to be rewiten
     */
    protected static String URLEncode(String path) {

        return URLUtil.URLEncode(path, "UTF8");

    }
    
    
}

